#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: auto_build_scorecard.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import os
import time
import warnings

import numpy as np
import pandas as pd
import xlsxwriter

import autobmt

warnings.filterwarnings('ignore')

log = autobmt.Logger(level='info', name=__name__).logger


class AutoBuildScoreCard:

    def __init__(self, datasets, fea_names, target, key='key', data_type='type',
                 no_feature_names=['key', 'target', 'apply_time', 'type'], ml_res_save_path='./model_result'):

        if data_type not in datasets:
            raise KeyError('train、test数据集标识的字段名不存在！或未进行数据集的划分，请将数据集划分为train、test！！！')

        datasets[data_type] = datasets[data_type].map(str.lower)

        self.data_type_ar = np.unique(datasets[data_type])
        if 'train' not in self.data_type_ar:
            raise KeyError("""没有开发样本，数据集标识字段{}没有`train`该取值！！！""".format(data_type))

        if 'test' not in self.data_type_ar:
            raise KeyError("""没有验证样本，数据集标识字段{}没有`test`该取值！！！""".format(data_type))

        if target not in datasets:
            raise KeyError('样本中没有目标变量y值！！！')

        # fea_names = [i for i in fea_names if i != key and i != target]
        fea_names = [i for i in fea_names if i not in no_feature_names]
        log.info('数据集变量个数 : {}'.format(len(fea_names)))
        log.info('fea_names is : {}'.format(fea_names))

        self.datasets = datasets
        self.fea_names = fea_names
        self.target = target
        self.key = key
        self.no_feature_names = no_feature_names
        self.ml_res_save_path = ml_res_save_path + '/' + time.strftime('%Y%m%d%H%M%S_%s', time.localtime())

        os.makedirs(self.ml_res_save_path, exist_ok=True)

    def fit(self, empty_threhold=0.9, iv_threhold=0.02, corr_threhold=0.7, psi_threhold=0.1,
            dev_nodev_iv_diff_threhold=0.08):
        log.info('样本数据集类型：{}'.format(self.data_type_ar))
        log.info('样本行列情况：{}'.format(self.datasets.shape))
        log.info('样本正负占比情况：')
        log.info(self.datasets[self.target].value_counts() / len(self.datasets))

        excel_file_name_path = "{}/build_model_log.xlsx".format(self.ml_res_save_path)
        workbook = xlsxwriter.Workbook(excel_file_name_path, {'nan_inf_to_errors': True})
        excel_utils = autobmt.Report2Excel(excel_file_name_path, workbook=workbook)
        log.info('读取样本&特征数据集：{} |为样本数据，其他为特征数据'.format(self.no_feature_names))
        # =========================读取数据集结束=========================

        time_start = time.time()

        log.info('Step 1: EDA，整体数据探索性数据分析')
        all_data_eda = autobmt.detect(self.datasets)
        all_data_eda.to_excel('{}/all_data_eda.xlsx'.format(self.ml_res_save_path))
        excel_utils.df_to_excel(all_data_eda, '1.EDA')

        log.info('Step 2: 特征粗筛选开始')
        # 进行特征初步选择
        # 分箱方法支持 dt、chi、equal_freq、kmeans
        fs_dic = {
            "empty": {'threshold': empty_threhold},
            # "const": {'threshold': 0.95},
            "psi": {'threshold': psi_threhold},
            "iv": {'threshold': iv_threhold},
            # "iv_diff": {'threshold': dev_nodev_iv_diff_threhold},
            "corr": {'threshold': corr_threhold},

        }

        # TODO 考虑将self.datasets更换为train_data
        fs = autobmt.FeatureSelection(df=self.datasets, target=self.target, exclude_columns=self.no_feature_names,
                                      match_dict=None,
                                      params=fs_dic)
        #selected_df, selected_features, select_log_df, selection_evaluate_log_df, fbfb = fs.select()
        selected_df, selected_features, select_log_df, fbfb = fs.select()
        summary = autobmt.calc_var_summary(
            fbfb.transform(selected_df[selected_df['type'] == 'train'])[selected_features + [self.target]],
            fbfb.export(), target=self.target, need_bin=False)
        # summary = fs.get_binning_summary  # 获取分箱详情
        autobmt.del_df(self.datasets)

        log.info('Step 2: 特征粗筛选结束')

        log.info('Step 3: 对剩下的变量调用最优分箱')
        log.info('剩下的变量个数 : '.format(len(selected_features)))
        log.info('剩下的变量 : '.format(selected_features))

        # train_data = selected_df[selected_df['type'] == 'train']

        # 训练集
        fb, best_binning_result = autobmt.best_binning(selected_df[selected_df['type'] == 'train'],
                                                       x_list=selected_features,
                                                       target=self.target)
        df_bin = fb.transform(selected_df, labels=False)
        train_bin = df_bin[df_bin['type'] == 'train']
        test_bin = df_bin[df_bin['type'] == 'test']
        if 'oot' in self.data_type_ar:
            oot_bin = df_bin[df_bin['type'] == 'oot']

        train_var_summary = autobmt.calc_var_summary(train_bin[selected_features + [self.target]], fb.export(),
                                                     target=self.target, need_bin=False)

        # 测试集
        # test_bin = fb.transform(test_data, labels=False)
        test_var_summary = autobmt.calc_var_summary(test_bin[selected_features + [self.target]], fb.export(),
                                                    target=self.target, need_bin=False)

        if 'oot' in self.data_type_ar:
            # oot_bin = fb.transform(oot_data, labels=False)
            oot_var_summary = autobmt.calc_var_summary(oot_bin[selected_features + [self.target]], fb.export(),
                                                       target=self.target, need_bin=False)

        # 计算拐点
        train_data_inflection_df = autobmt.compare_inflection_point(train_var_summary)
        train_data_inflection_df.columns = train_data_inflection_df.columns.map(lambda x: 'train_' + x)

        test_data_inflection_df = autobmt.compare_inflection_point(test_var_summary)
        test_data_inflection_df.columns = test_data_inflection_df.columns.map(lambda x: 'test_' + x)
        data_inflection_df = pd.concat([train_data_inflection_df, test_data_inflection_df], axis=1)
        if 'oot' in self.data_type_ar:
            oot_data_inflection_df = autobmt.compare_inflection_point(oot_var_summary)
            oot_data_inflection_df.columns = oot_data_inflection_df.columns.map(lambda x: 'oot_' + x)
            data_inflection_df = pd.concat([data_inflection_df, oot_data_inflection_df], axis=1)

        # 存储特征粗筛内容
        # 存储最优分箱内容
        excel_utils.df_to_excel(select_log_df, '2.feature_selection')
        autobmt.var_summary_to_excel(summary, workbook, '3.binning_summary')
        autobmt.var_summary_to_excel(train_var_summary, workbook, '4.best_train_binning_summary')
        autobmt.var_summary_to_excel(test_var_summary, workbook, '5.best_test_binning_summary')
        if 'oot' in self.data_type_ar:
            autobmt.var_summary_to_excel(oot_var_summary, workbook, '6.best_oot_binning_summary')
        excel_utils.df_to_excel(best_binning_result, '7.best_binning_log')
        excel_utils.df_to_excel(data_inflection_df.reset_index(), '8.data_inflection_point')
        # 转woe
        log.info('Step 4: 对剩下的变量进行woe转换')
        woetf = autobmt.WoeTransformer()
        train_woe = woetf.fit_transform(train_bin, train_bin[self.target], exclude=self.no_feature_names)
        test_woe = woetf.transform(test_bin)
        if 'oot' in self.data_type_ar:
            oot_woe = woetf.transform(oot_bin)

        # if 'oot' in self.data_type_ar:
        #     bin_data = pd.concat([train_bin, test_bin, oot_bin])
        # else:
        #     bin_data = pd.concat([train_bin, test_bin])
        # dump_model_to_file(bin_data, '{}/bin_data.pkl'.format(self.ml_res_save_path))

        #log.info('Step 5: 获取变量的iv和psi')

        # TODO list：考虑每月和总体的psi比较
        # # 计算变量的iv和psi以及missing rate
        # psi_v, frame = psi(train_woe[selected_features], test_woe[selected_features], return_frame=True)
        # iv_psi_df = pd.DataFrame(train_var_summary.groupby('var_name')['IV'].mean()).merge(pd.DataFrame(psi_v),
        #                                                                                    left_index=True,
        #                                                                                    right_index=True).rename(
        #     columns={0: 'PSI'})
        # select_log_df.set_index('feature', inplace=True)
        # iv_psi_missing_df = select_log_df[select_log_df.index.isin(selected_features)][['miss_rate']].merge(iv_psi_df,
        #                                                                                                     how='right',
        #                                                                                                     left_index=True,
        #                                                                                                     right_index=True)

        # TODO list：iv_psi_missing_df再加上coef、pvalue、vif
        # TODO list：1、通过稳定性筛选特征 2、由于分箱转woe值后，变量之间的共线性会变强，通过相关性再次筛选特征

        log.info('Step 5: 对woe转换后的变量进行stepwise')
        if 'oot' in self.data_type_ar:
            in_model_data = pd.concat([train_woe, test_woe, oot_woe])
        else:
            in_model_data = pd.concat([train_woe, test_woe])

        # dump_model_to_file(in_model_data, '{}/in_model_data.pkl'.format(self.ml_res_save_path))
        var_bin_woe = autobmt.fea_woe_dict_format(woetf.export(), fb.export())
        # sw = StepWise(in_model_data, self.target, selected_features, self.no_feature_names, iv_psi_missing_df,
        #               var_bin_woe, {})
        # selected_features, stepwise_df, evaluate_df, stepwise_evaluate_log_df = sw.stepwise_apply()
        # # selected_features, stepwise_df,stepwise_log_df, evaluate_df, stepwise_evaluate_log_df = sw.run()

        # 将woe转化后的数据做逐步回归
        final_data = autobmt.stepwise(train_woe, target=self.target, estimator='ols', direction='both',
                                      criterion='aic',
                                      exclude=self.no_feature_names)

        # # 将选出的变量应用于test/oot数据
        # final_test = test_woe[final_data.columns]
        # if 'oot' in self.data_type_ar:
        #     final_oot = oot_woe[final_data.columns]
        # print(final_data.shape)  # 逐步回归从31个变量中选出了10个

        # 确定建模要用的变量
        selected_features = [fea for fea in final_data.columns if fea not in self.no_feature_names]

        log.info('Step 6: 用逻辑回归构建模型')
        # 用逻辑回归建模
        from sklearn.linear_model import LogisticRegression

        # lr = LogisticRegression()
        # lr.fit(final_data[selected_features], final_data[self.target])
        while True:  # 循环的目的是保证入模变量的系数都为整
            lr = LogisticRegression()
            lr.fit(final_data[selected_features], final_data[self.target])
            drop_var = np.array(selected_features)[np.where(lr.coef_ < 0)[1]]
            if len(drop_var) == 0:
                break
            selected_features = list(set(selected_features) - set(drop_var))

        # # 预测训练和隔月的OOT
        # pred_train = lr.predict_proba(final_data[selected_features])[:, 1]
        # pred_test = lr.predict_proba(final_test[selected_features])[:, 1]
        # if 'oot' in self.data_type_ar:
        #     pred_oot = lr.predict_proba(final_oot[selected_features])[:, 1]
        # # 预测训练和隔月的OOT

        in_model_data['p'] = lr.predict_proba(in_model_data[selected_features])[:, 1]
        # dump_model_to_file(in_model_data[self.no_feature_names + selected_features + ['p']],
        #                    '{}/lr_in_model_data.pkl'.format(self.ml_res_save_path))

        ###
        psi_v = autobmt.psi(test_woe[selected_features], train_woe[selected_features])
        psi_v.name = 'train_test_psi'
        train_iv = train_var_summary[['var_name', 'IV']].rename(columns={'IV': 'train_iv'}).drop_duplicates().set_index(
            'var_name')
        test_iv = train_var_summary[['var_name', 'IV']].rename(columns={'IV': 'test_iv'}).drop_duplicates().set_index(
            'var_name')
        var_miss = select_log_df[['feature', 'cn', 'miss_rate']].drop_duplicates().set_index('feature')

        if 'oot' in self.data_type_ar:
            psi_o = autobmt.psi(oot_woe[selected_features], train_woe[selected_features])
            psi_o.name = 'train_oot_psi'
            oot_iv = train_var_summary[['var_name', 'IV']].rename(columns={'IV': 'oot_iv'}).drop_duplicates().set_index(
                'var_name')

        coef_s = {}
        for idx, key in enumerate(selected_features):
            coef_s[key] = lr.coef_[0][idx]
        var_coef = pd.Series(coef_s, name='coef')
        var_vif = autobmt.get_vif(final_data[selected_features])
        statsm = autobmt.StatsModel(estimator='ols', intercept=True)
        t_p_c_value = statsm.stats(final_data[selected_features], final_data[self.target])
        p_value = pd.Series(t_p_c_value['p_value'], name='p_value')
        t_value = pd.Series(t_p_c_value['t_value'], name='t_value')

        if 'oot' in self.data_type_ar:
            var_info = pd.concat(
                [var_coef, p_value, t_value, var_vif, train_iv, test_iv, oot_iv, psi_v, psi_o, var_miss],
                axis=1).dropna(subset=['coef'])
        else:
            var_info = pd.concat([var_coef, p_value, t_value, var_vif, train_iv, test_iv, psi_v, var_miss],
                                 axis=1).dropna(subset=['coef'])
        ###

        log.info('Step 7: 构建评分卡')
        card = autobmt.ScoreCard(
            combiner=fb,
            transer=woetf,
            # class_weight = 'balanced',
            # C=0.1,
            # base_score = 600,
            # base_odds = 15 ,
            # pdo = 50,
            # rate = 2
        )

        card.fit(final_data[selected_features], final_data[self.target])

        log.info('Step 8: 持久化模型，分箱点，woe值，评分卡结构======>开始')
        autobmt.dump_to_pkl(fb, '{}/fb.pkl'.format(self.ml_res_save_path))
        autobmt.dump_to_pkl(woetf, '{}/woetf.pkl'.format(self.ml_res_save_path))
        autobmt.dump_to_pkl(selected_features, '{}/in_model_var.pkl'.format(self.ml_res_save_path))

        woetf.export(to_json='{}/var_bin_woe.json'.format(self.ml_res_save_path))
        woetf.export(to_json='{}/var_bin_woe_format.json'.format(self.ml_res_save_path), var_bin_woe=var_bin_woe)
        fb.export(to_json='{}/var_split_point.json'.format(self.ml_res_save_path), bin_format=False)
        fb.export(to_json='{}/var_split_point_format.json'.format(self.ml_res_save_path))
        card.export(to_json='{}/scorecard.json'.format(self.ml_res_save_path))

        woetf.export(to_csv='{}/var_bin_woe.csv'.format(self.ml_res_save_path))
        woetf.export(to_csv='{}/var_bin_woe_format.csv'.format(self.ml_res_save_path), var_bin_woe=var_bin_woe)
        fb.export(to_csv='{}/var_split_point.csv'.format(self.ml_res_save_path), bin_format=False)
        fb.export(to_csv='{}/var_split_point_format.csv'.format(self.ml_res_save_path))
        scorecard_structure = card.export(to_dataframe=True)
        scorecard_structure.to_csv('{}/scorecard.csv'.format(self.ml_res_save_path), index=False)

        # dump_model_to_file(sw.model, '{}/lr_sm_model.pkl'.format(self.ml_res_save_path))
        # output_model_data = sw.get_output_data
        # scorecard_structure = sw.get_scorecard_structure()

        autobmt.dump_to_pkl(lr, '{}/lrmodel.pkl'.format(self.ml_res_save_path))
        autobmt.dump_to_pkl(card, '{}/scorecard.pkl'.format(self.ml_res_save_path))
        in_model_data['score'] = in_model_data['p'].map(autobmt.to_score)
        output_report_data = in_model_data[self.no_feature_names + ['p', 'score']]
        # in_model_data['card_score'] = card.predict(self.datasets)
        # output_report_data = in_model_data[self.no_feature_names + ['p', 'score', 'card_score']]

        # dump_model_to_file(output_report_data, '{}/output_report_data.pkl'.format(self.ml_res_save_path))
        output_report_data.to_csv('{}/lr_pred_to_report_data.csv'.format(self.ml_res_save_path), index=False)  #
        selected_df[self.no_feature_names + selected_features].head(500).to_csv(
            '{}/lr_test_input.csv'.format(self.ml_res_save_path),
            index=False)
        lr_auc_ks_psi = autobmt.get_auc_ks_psi(output_report_data)
        lr_auc_ks_psi.to_csv('{}/lr_auc_ks_psi.csv'.format(self.ml_res_save_path),
                                                          index=False)
        # stepwise_df.reset_index().to_csv('{}/feature_coef_pvalue_IV_vif_psi_corr.csv'.format(self.ml_res_save_path),
        #                                  index=False)
        # dump_model_to_file(scorecard_structure, '{}/scorecard_structure.pkl'.format(self.ml_res_save_path))
        log.info('Step 8: 持久化模型，分箱点，woe值，评分卡结构======>结束')

        log.info('Step 9: 持久化建模中间结果到excel，方便复盘')
        # corr
        train_woe_corr_df = train_woe[selected_features].corr().reset_index()
        test_woe_corr_df = test_woe[selected_features].corr().reset_index()
        if 'oot' in self.data_type_ar:
            oot_woe_corr_df = oot_woe[selected_features].corr().reset_index()

        # 建模中间结果存储
        # 将模型最终评估结果合并到每一步的评估中

        # final_evaluate_df = merge_rows_one_row_df(evaluate_df, name='lr_', stepname='scorecard_evaluate')
        # final_step_evaluate_log_df = pd.concat([selection_evaluate_log_df, stepwise_evaluate_log_df, final_evaluate_df],
        #                                        axis=0).reset_index(drop=True)
        #
        # excel_utils.df_to_excel(final_step_evaluate_log_df, '9.step_evaluate')  # 每步筛选的评估结果
        # excel_utils.df_to_excel(stepwise_df.reset_index(), '10.model_stepwise')  # stepwise模型结果
        # sw.scorecard_to_excel(scorecard_structure, workbook, 'scorecard_structure')
        excel_utils.df_to_excel(var_info.reset_index().rename(columns={'index': 'var_name'}), '9.var_info')
        excel_utils.df_to_excel(scorecard_structure, 'scorecard_structure')

        in_model_var_train_summary = train_var_summary[train_var_summary['var_name'].isin(selected_features)]
        in_model_var_test_summary = test_var_summary[test_var_summary['var_name'].isin(selected_features)]
        if 'oot' in self.data_type_ar:
            in_model_var_oot_summary = oot_var_summary[oot_var_summary['var_name'].isin(selected_features)]
        autobmt.var_summary_to_excel(in_model_var_train_summary, workbook, '11.in_model_var_train_summary')
        autobmt.var_summary_to_excel(in_model_var_test_summary, workbook, '12.in_model_var_test_summary')
        if 'oot' in self.data_type_ar:
            autobmt.var_summary_to_excel(in_model_var_oot_summary, workbook, '13.in_model_var_oot_summary')

        excel_utils.df_to_excel(train_woe_corr_df, '14.train_woe_df_corr')
        excel_utils.df_to_excel(test_woe_corr_df, '15.test_woe_df_corr')
        if 'oot' in self.data_type_ar:
            excel_utils.df_to_excel(oot_woe_corr_df, '16.oot_woe_df_corr')
        excel_utils.close_workbook()

        if 'oot' in self.data_type_ar:
            autobmt.plot_var_bin_summary(
                {'train': in_model_var_train_summary, 'test': in_model_var_test_summary,
                 'oot': in_model_var_oot_summary},
                selected_features, target=self.target, by=self.data_type_ar, file_path=excel_file_name_path,
                sheet_name='plot_var_bin_summary')
        else:
            autobmt.plot_var_bin_summary(
                {'train': in_model_var_train_summary, 'test': in_model_var_test_summary},
                selected_features, target=self.target, by=self.data_type_ar, file_path=excel_file_name_path,
                sheet_name='plot_var_bin_summary')

        time_end = time.time()
        time_c = time_end - time_start
        log.info('*' * 30 + '建模相关结果保存完成！！！保存路径为：{}'.format(self.ml_res_save_path) + '*' * 30)
        log.info('模型效果：{}'.format(lr_auc_ks_psi))
        log.info('time cost {} s'.format(time_c))

        return lr, selected_features

    @classmethod
    def predict(cls, to_pred_df=None, model_path=None):
        if to_pred_df is None:
            raise ValueError('需要进行预测的数据集不能为None，请指定数据集！！！')
        if model_path is None:
            raise ValueError('模型路径不能为None，请指定模型文件路径！！！')
        fb = autobmt.load_from_pkl('{}/fb.pkl'.format(model_path))
        woetf = autobmt.load_from_pkl('{}/woetf.pkl'.format(model_path))
        lrmodel = autobmt.load_from_pkl('{}/lrmodel.pkl'.format(model_path))
        selected_features = autobmt.load_from_pkl('{}/in_model_var.pkl'.format(model_path))

        bin_data = fb.transform(to_pred_df)
        woe_data = woetf.transform(bin_data)
        to_pred_df['p'] = lrmodel.predict_proba(woe_data[selected_features])[:, 1]

        return to_pred_df



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: statistics.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-03
'''

import re

import numpy as np
import pandas as pd
import statsmodels.api as sm
from joblib import Parallel, delayed

import autobmt
from .metrics import psi
from .utils import is_continuous, to_ndarray, np_count, support_dataframe, split_points_to_bin


# TODO compare_inflection_point和calc_iv_and_inflection_point考虑要合并

def compare_inflection_point(df_summary):
    """
    比较数据集的拐点趋势
    切记，数据集使用type区分，因此一定要包含type列，数据集是分好箱的数据集，不是原始数据集而是转woe后的
    因为需要转换计算badrate
    Args:
        df_summary (DataFrame):分好箱的数据集，可能包含train/test/oot

    Returns:
        DataFrame: 拐点信息
    """

    def __calc_group(data):
        range = data.range.tolist()
        nan_bin = '{}.nan'.format(len(range) - 1)
        if nan_bin in range:
            data = data[~data['range'].isin([nan_bin])]
        badrate = data.positive_rate
        is_monotonic = badrate.is_monotonic_decreasing or badrate.is_monotonic_increasing

        inflection_point_index, inflection_shape = get_inflection_point_index(badrate)
        return pd.Series({'is_monotonic': is_monotonic, 'bin_count': len(range),
                          'inflection_point_num': len(inflection_point_index),
                          'inflection_point_index': inflection_point_index,
                          'inflection_shape': inflection_shape})

    data_inflection_df = df_summary.groupby('var_name').apply(__calc_group)
    return data_inflection_df


def merge_rows_one_row_df(df, name="", stepname=None):
    """将一个多行的dataframe合并成只有一行的dataframe"""
    tmp_arr = []
    for i in range(df.shape[0]):
        tmp = df.iloc[i, :].add_prefix("{}_{}".format(df.index[i], name))
        tmp_arr.append(tmp)

    result_df = pd.DataFrame(pd.concat(tmp_arr, axis=0)).T
    if stepname is not None:  # 合并成一行后,增加一列标识，用于和别的评估进行区分
        result_df['stepname'] = stepname
    return result_df


def calc_iv_and_inflection_point(df, target='target', bin_func=None, bin_format={}):
    """
    计算iv，拐点
    Args:
        df (DataFrame):分好箱转换后的数据集
        target (str):目标变量
        bin_func (str): 分箱方法
        bin_format (dict): 格式化好的分箱点

    Returns:
        DataFrame
    """

    def __calc_group(data, bin_func):
        range = data.range.tolist()
        nan_bin = '{}.nan'.format(len(range) - 1)
        if nan_bin in range:
            data = data[~data['range'].isin([nan_bin])]
        bin_badrate = data.positive_rate
        inflection_point_index, inflection_shape = get_inflection_point_index(bin_badrate)
        return pd.Series({'IV': data.IV.get(0), 'bin_count': len(range),
                          'inflection_point_num': len(inflection_point_index),
                          'inflection_point_index': inflection_point_index,
                          'inflection_shape': inflection_shape, 'bin_func': bin_func})

    summary = Parallel(n_jobs=-1)(
        delayed(calc_bin_summary)(df[[col, target]], bin_col=col, bin=False, target=target, is_sort=False,
                                  bin_format=bin_format) for col in df.columns if col not in [target])
    var_summary = pd.concat(summary, axis=0)

    var_iv_inflection_df = var_summary.groupby('var_name').apply(__calc_group, bin_func)
    return var_iv_inflection_df


def calc_var_summary(df, bin_format={}, include_cols=[], target='target', need_bin=True, **kwargs):
    """
    计算所有变量的详情
    Args:
        df (DataFrame): 含有目标变量及分箱后的数据集
        bin_format (dict): 格式化好的分箱点
        include_cols (list): 需要统计的特征
        target (str): 目标值变量名称
        need_bin (bool): 是否需要分箱

    Returns:
        DataFrame: 变量分箱详情
    """

    if include_cols:
        cols = np.array(include_cols)
    else:
        cols = np.array(df.columns)

    if not need_bin:
        kwargs = {'bin': False}
    summary = Parallel(n_jobs=-1)(
        delayed(calc_bin_summary)(df[[col, target]], bin_col=col, target=target, is_sort=False, bin_format=bin_format,
                                  **kwargs) for col in cols if col != target)
    var_summary = pd.concat(summary, axis=0)
    return var_summary


def calc_bin_summary(df, bin_col='score', bin=10, target='target', is_sort=True, method='equal_freq',
                     is_need_monotonic=False, bin_format={}, **kwargs):
    """
    变量分箱详情
    Args:
        df (DataFrame): 含目标变量的数据集
        bin_col (str): 需要计算的列名
        bin (int): 分几箱
        target (str):目标变量列名
        method (str):分箱方法；'dt'、'chi'、'equal_freq'、'kmeans'四种供选择
        is_sort (bool):是否需要排序，默认True，倒序排序
        **kwargs:

    Returns:
        DataFrame: 变量分箱详情
    """

    # negative:good，positive:bad
    def __calc_group(data_, var_name):
        """获取分组的人数,好坏人数"""
        count = len(data_)
        bad_num = data_.y.sum()
        good_num = count - bad_num
        return pd.Series(
            {'var_name': var_name, 'min': min(data_.x), 'max': max(data_.x), 'positive': bad_num, 'negative': good_num,
             'total': count})

    data = pd.DataFrame({'x': df[bin_col], 'y': df[target]})
    total = len(data)
    positive_count = data.y.sum()
    negative_count = total - positive_count

    data['range'] = 0
    if bin is False:
        data['range'] = data.x
    elif isinstance(bin, (list, np.ndarray, pd.Series)):
        if len(bin) < len(data.x):
            bin = split_points_to_bin(data.x, bin)

        data['range'] = bin
    elif isinstance(bin, int):
        fb = autobmt.FeatureBin()
        fb.fit(data.x, data.y, n_bins=bin, method=method, is_need_monotonic=is_need_monotonic, **kwargs)
        data['range'] = fb.transform(data.x)
        bin_format = fb.export()
        bin_format[bin_col] = bin_format.pop('x')

    bin_g = data.groupby(data['range'], dropna=False).apply(__calc_group, var_name=bin_col)

    if is_sort:
        bin_g.sort_values(by='min', ascending=False, inplace=True)  # 正常

    bin_g['positive_rate'] = bin_g['positive'] / bin_g['total']  # bad_rate,,,区间_正样本比率
    bin_g['negative_rate'] = bin_g['negative'] / bin_g['total']  # 区间_负样本比率

    bin_g['odds'] = bin_g['positive'] / bin_g['negative']

    # bin_g['positive_pct'] = bin_g['positive'] / positive_count  # 区间正样本占比
    bin_g['positive_pct'] = bin_g['positive'].map(
        lambda x: 1 / positive_count if x == 0 else x / positive_count)  # 区间正样本占比
    # bin_g['negative_pct'] = bin_g['negative'] / negative_count  # 区间负样本/总体负样本
    bin_g['negative_pct'] = bin_g['negative'].map(
        lambda x: 1 / negative_count if x == 0 else x / negative_count)  # 区间负样本/总体负样本
    bin_g['total_pct'] = bin_g['total'] / total  # 区间总人数/总体总人数

    bin_g['cum_negative_pct'] = bin_g['negative_pct'].cumsum()  # 累计负样本人数占比

    cum_positive = bin_g['positive'].cumsum()  # 累计正样本人数
    cum_total = bin_g['total'].cumsum()  # 累计总人数

    bin_g['cum_total_pct'] = cum_total / total  # 累计通过人数占比，累计捕获率，agg2['cum_total_prop'] = cum_total / all_total
    bin_g['cum_positive_pct'] = bin_g['positive_pct'].cumsum()  # 查全率,,,累计正样本人数占比
    bin_g['cum_positive_rate'] = cum_positive / cum_total  # 查准率,,,累计捕获的样本中正样本的占比

    bin_g['ks'] = bin_g['cum_positive_pct'] - bin_g['cum_negative_pct']  # 累计正样本人数占比/累计负样本人数占比

    bin_g['lift'] = bin_g['positive_rate'] / (positive_count / total)
    if bin_g['ks'].sum() < 0:
        bin_g['ks'] = -bin_g['ks']
        bin_g['cum_negative_pct'] = bin_g.loc[::-1, 'negative_pct'].cumsum()[::-1]  # 累计负样本人数占比
        cum_positive_rev = bin_g.loc[::-1, 'positive'].cumsum()[::-1]  # 累计正样本人数
        cum_total_rev = bin_g.loc[::-1, 'total'].cumsum()[::-1]  # 累计总人数
        bin_g['cum_total_pct'] = cum_total_rev / total  # 累计通过人数占比，累计捕获率，agg2['cum_total_prop'] = cum_total / all_total
        bin_g['cum_positive_pct'] = bin_g.loc[::-1, 'positive_pct'].cumsum()[::-1]  # 查全率,,,累计正样本人数占比
        bin_g['cum_positive_rate'] = cum_positive_rev / cum_total_rev  # 查准率,,,累计捕获的样本中正样本的占比

    bin_g['cum_lift'] = bin_g['cum_positive_pct'] / bin_g['cum_total_pct']

    bin_g['woe'] = bin_g.apply(lambda x: np.log(x['positive_pct'] / x['negative_pct']), axis=1)
    bin_g['iv'] = (bin_g['positive_pct'] - bin_g['negative_pct']) * bin_g.woe
    bin_g['IV'] = bin_g.iv.sum()

    bin_g.index.name = 'range'
    bin_g = bin_g.reset_index()

    if bin_col in bin_format:
        range_format = {int(re.match(r"^(\d+)\.", i).group(1)): i for i in bin_format[bin_col]}
        bin_g['range_num'] = bin_g['range']
        bin_g['range'] = bin_g['range'].map(range_format)
    else:
        bin_g['range_num'] = bin_g['range'].fillna(len(bin_g))

    return bin_g


def calc_woe_iv(df, col_name='default_name', bin_format={}, target='target'):
    """
    已对齐IV的计算方式
    计算单变量详情，woe,iv值
    Args:
        df: 含有目标变量及分箱后的数据集
        col_name: 单变量名称
        bin_format: 格式化好的分箱点
        target: 目标值变量名称

    Returns:

    """

    def __calc_group(data_, var_name):
        """获取分组的人数,好坏人数"""
        count = len(data_)
        bad_num = data_.Y.sum()
        good_num = count - bad_num

        return pd.Series({'var_name': var_name, 'Total': count, 'Bad': bad_num, 'Good': good_num})

    X, Y = df[col_name], df[target]

    data = pd.DataFrame({'X': X, 'Y': Y})

    bin_g = data.groupby(data['X'], dropna=False).apply(__calc_group, var_name=col_name)
    total = data.Y.count()
    bad_count = (data.Y == 1).sum()
    good_count = (data.Y == 0).sum()
    bin_g['Bad_Rate'] = bin_g['Bad'] / bin_g['Total']  # bad_rate
    # bin_g['Pct_Bad'] = bin_g['Bad'] / bad_count  # bad_人数占比
    # bin_g['Pct_Good'] = bin_g['Good'] / good_count  # good_人数占比
    bin_g['Pct_Bad'] = bin_g['Bad'].map(lambda x: 1 / bad_count if x == 0 else x / bad_count)  # bad_人数占比
    bin_g['Pct_Good'] = bin_g['Good'].map(lambda x: 1 / good_count if x == 0 else x / good_count)  # good_人数占比
    bin_g['Pct_Bin'] = bin_g['Total'] / total  # 总人数占比
    # bin_g['累计坏人数'] = bin_g['Good'].cumsum()
    # bin_g['累计好人数'] = bin_g['Good'].cumsum()
    # bin_g['累计坏人数占比'] = bin_g['Good'].cumsum() / bad_count
    # bin_g['累计好人数占比'] = bin_g['Good'].cumsum() / good_count
    # bin_g['woe'] = bin_g.apply(
    #     lambda x: 0.0 if x['Pct_Good'] == 0 or x['Pct_Bad'] == 0 else round(np.log(x['Pct_Bad'] / x['Pct_Good']),
    #                                                                         5),
    #     axis=1)
    bin_g['woe'] = bin_g.apply(lambda x: np.log(x['Pct_Bad'] / x['Pct_Good']), axis=1)
    # bin_g['ks'] = abs(bin_g['累计坏人数占比'] - bin_g['累计好人数占比'])
    bin_g['iv'] = (bin_g['Pct_Bad'] - bin_g['Pct_Good']) * bin_g.woe
    bin_g['IV'] = bin_g.iv.replace({np.inf: 0, -np.inf: 0}).sum()
    bin_g.index.name = 'range'
    bin_g = bin_g.reset_index()
    if col_name in bin_format:
        range_format = {int(re.match(r"^(\d+)\.", i).group(1)): i for i in bin_format[col_name]}
        bin_g['range_num'] = bin_g['range']
        bin_g['range'] = bin_g['range'].map(range_format)
    else:
        bin_g['range_num'] = bin_g['range'].fillna(len(bin_g))

    # 每个分箱字段之间加上一个空行
    # bin_g = bin_g.append(pd.Series(),ignore_index=True)
    return bin_g


def get_inflection_point_index(arr):
    """
    返回一个数组的拐点索引,以及单调情况
    Args:
        arr (array) : 数组

        + 单调递增
        - 单调递减
        u u形曲线
        ~u 倒u形曲线
        ~ 不单调
    Returns:
        array: 拐点的位置
        str: 单调标志
    """
    diff_arr = np.diff(arr).tolist()
    # 返回是单调递增(+)，还是单调递减(-)，还是不单调(~)
    monotonic_flag = check_monotonic(diff_arr)
    index_arr = []  # 记录拐点的位置
    for i in range(0, len(diff_arr) - 1):
        if np.signbit(diff_arr[i]) != np.signbit(diff_arr[i + 1]):
            index_arr.append(i + 1)
    if len(index_arr) == 1:
        monotonic_flag = "~U" if arr[1] - arr[0] > 0 else "U"

    return index_arr, monotonic_flag


def check_monotonic(arr):
    """判断是单调递增还是单调递减,先调用arr = np.diff(arr)"""
    count = np.sum(np.array(arr) > 0)
    if count == len(arr):
        return "+"  # 单调递增
    elif count == 0:
        return "-"  # 单调递减
    else:
        return "~"  # 不单调


def get_vif(frame):
    """
    计算VIF
    Args:
        frame:  (ndarray|DataFrame)

    Returns:
        Series
    """
    index = None
    if isinstance(frame, pd.DataFrame):
        index = frame.columns
        frame = frame.values

    l = frame.shape[1]
    vif = np.zeros(l)

    for i in range(l):
        X = frame[:, np.arange(l) != i]
        y = frame[:, i]

        model = sm.OLS(y, X)
        r2 = model.fit().rsquared_adj

        vif[i] = 1 / (1 - r2)

    return pd.Series(vif, index=index, name='vif')


def WOE(y_prob, n_prob):
    """计算woe值

    Args:
        y_prob: 正样本在整个正样本中的占比
        n_prob: 负样本在整个负样本中的占比

    Returns:
        number: woe 值
    """
    return np.log(y_prob / n_prob)


def _IV(feature, target):
    """计算IV值

    Args:
        feature (array-like)
        target (array-like)

    Returns:
        number: IV值
    """
    feature = to_ndarray(feature, dtype='str')
    target = to_ndarray(target)

    value = 0

    for v in np.unique(feature):
        y_prob, n_prob = probability(target, mask=(feature == v))
        value += (y_prob - n_prob) * WOE(y_prob, n_prob)

    return value


@support_dataframe()
def calc_iv(feature, target, feature_bin=None, **kwargs):
    """计算1个特征的IV值

    Args:
        feature (array-like)
        target (array-like)
        n_bins (int): 需要分几箱
        method (str): 分箱方法；'dt'、'chi'、'equal_freq'、'kmeans'四种供选择， 默认 'dt'
        **kwargs (): bin_method_run分箱函数的其它参数
    """

    ###TODO: 考虑增加是否单调的参数

    if is_continuous(feature):

        if feature_bin is not None:
            if hasattr(feature, 'name') and feature.name in feature_bin.splits_dict:
                feature = feature_bin.transform(feature)
            else:
                feature = feature_bin.fit_transform(feature, target, method='dt', is_need_monotonic=False, **kwargs)
        else:
            if 'return_bin' in kwargs: del kwargs['return_bin']
            s, feature = autobmt.bin_method_run(feature, target, return_bin=True, is_need_monotonic=False, **kwargs)

    return _IV(feature, target)


def bin_badrate(feature, target=None):
    """
    计算badrate【即正样本占比】
    Args:
        feature:
        target:

    Returns:

    """
    badrate_list = []
    bin_rate_list = []
    total = len(feature)
    uni_fea = np.sort(np.unique(feature))
    # uni_fea = np.unique(feature)
    if target is None:
        for value in uni_fea:
            # mask = (feature == value)
            mask = feature == value
            bin_rate_list.append(np.sum(mask) / total)
        return min(bin_rate_list)
    else:
        for value in uni_fea:
            # mask = (feature == value)
            mask = feature == value
            bin_target = target[mask]
            bin_badrate = np.sum(bin_target) / len(bin_target)
            bin_rate_list.append(len(bin_target) / total)
            # bin_rate_list.append(np.sum(mask) / total)
            badrate_list.append(bin_badrate)
        return badrate_list, min(bin_rate_list)


def bin_badratebase(feature, target):
    badrate_list = []
    bin_rate_list = []
    total = len(feature)
    uni_fea = np.sort(np.unique(feature))
    # uni_fea = np.unique(feature)
    for value in uni_fea:
        # mask = (feature == value)
        mask = feature == value
        bin_target = target[mask]
        bin_badrate = np.sum(bin_target) / len(bin_target)
        bin_rate_list.append(len(bin_target) / total)
        badrate_list.append(bin_badrate)

    return badrate_list, min(bin_rate_list)


def probability(target, mask=None):
    """计算目标变量占比
    """
    if mask is None:
        return 1, 1

    counts_0 = np_count(target, 0, default=1)
    counts_1 = np_count(target, 1, default=1)

    sub_target = target[mask]

    sub_0 = np_count(sub_target, 0, default=1)
    sub_1 = np_count(sub_target, 1, default=1)

    y_prob = sub_1 / counts_1
    n_prob = sub_0 / counts_0

    return y_prob, n_prob


def get_iv_psi(df, feature_list=[], target='target', by_col='apply_mon', only_psi=True):
    """
    计算iv、psi
    Args:
        df (DataFrame): 原始数据集
        feature_list (list): 需要计算iv、psi的变量
        target (str): 目标变量名称
        by_col (str): 根据哪个变量分组计算iv、psi
        only_psi (bool): 是否只计算psi

    Returns:
        DataFrame: 变量的psi、iv
    """
    fb = autobmt.FeatureBin()
    fb.fit(df[feature_list], df[target], method='equal_freq', is_need_monotonic=False)
    # fb.fit(df[feature_list], df[target], method='dt')
    dev = fb.transform(df)
    by_col_v = sorted(list(set(df[by_col])))

    month_IV = pd.DataFrame()
    month_PSI_lis = []
    for n, j in enumerate(by_col_v):
        by_col_d = dev[dev[by_col] == j]
        if not only_psi:
            ###计算IV
            iv = {}
            iv_all = {}
            for i in feature_list:
                iv[i] = calc_iv(by_col_d[i], target=by_col_d[target], feature_bin=fb)
                if n == 0:
                    iv_all[i] = calc_iv(dev[i], target=dev[target], feature_bin=fb)
            if iv_all:
                month_IV = pd.concat([month_IV, pd.DataFrame([iv_all]).T.rename(columns={0: 'IV'})], axis=1)
            month_IV = pd.concat([month_IV, pd.DataFrame([iv]).T.rename(columns={0: f"{j}_IV"})], axis=1)
        ###计算PSI
        # by_col_psi = psi(dev[feature_list], by_col_d[feature_list])
        by_col_psi = psi(by_col_d[feature_list], dev[feature_list])
        by_col_psi.name = f"{j}_PSI"
        month_PSI_lis.append(by_col_psi)
    month_PSI = pd.DataFrame(month_PSI_lis).T
    month_PSI['MaxPSI'] = month_PSI.max(axis=1)

    ###iv趋势
    if not month_IV.empty:
        s_col_n = [f"{i}_IV" for i in by_col_v]
        for i in feature_list:
            month_IV.loc[i, 'IV趋势'] = get_inflection_point_index(month_IV.loc[i, s_col_n])[1]

    res = pd.concat([month_IV, month_PSI], axis=1)
    return res.sort_values(by='MaxPSI', ascending=False, )



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: feature_selection.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import numpy as np
import pandas as pd

from .metrics import psi
from .statistics import calc_iv
from .transformer import FeatureBin
from .utils import unpack_tuple, select_features_dtypes, is_continuous


def select_features_by_miss(df, nan=None, threshold=0.9, include_cols=[], return_drop=False, only_return_drop=False,
                            if_select_flow=False):
    """
    通过缺失率筛选特征
    Args:
        df (DataFrame): 需要进行特征筛选的数据集
        nan (str, regex, list, dict, Series, int, float, or None): 要替换为空的具体值
        threshold (float): 缺失率筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if nan is not None:
        df = df.replace(nan, np.nan)

    if include_cols:
        cols = include_cols
    else:
        cols = list(df.columns)

    missing_series = df[cols].isnull().sum() / len(df)

    del_c = list(missing_series[missing_series > threshold].index)

    if if_select_flow:
        return (del_c, threshold,
                pd.DataFrame({'feature': cols, 'miss_rate': missing_series}))  # TODO 需要检查下cols和psi_series是不是一一对应的

    if only_return_drop:
        return del_c

    r = df.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    return unpack_tuple(res)


def select_features_by_concentrate(df, nan=None, threshold=0.9, include_cols=[], return_drop=False,
                                   only_return_drop=False, if_select_flow=False):
    """
    通过集中度筛选特征
    Args:
        df (DataFrame): 需要进行特征筛选的数据集
        nan (str, regex, list, dict, Series, int, float, or None): 要替换为空的具体值
        threshold (float): 集中度筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if nan is not None:
        df = df.replace(nan, np.nan)

    if include_cols:
        cols = include_cols
    else:
        cols = list(df.columns)
    del_c = []
    max_ratios_ls = []
    row_count = df.shape[0]
    for col in cols:
        max_ratios = max(df[col].value_counts() / row_count)  # 缺失的占比不会被放进来
        # max_ratios = max(df[col].value_counts(dropna=False, normalize=True))    #缺失的占比会被放进来

        max_ratios_ls.append(max_ratios)
        if max_ratios > threshold:
            del_c.append(col)

    if if_select_flow:
        return (del_c, threshold,
                pd.DataFrame(
                    {'feature': cols, 'concentration_rate': max_ratios_ls}))

    if only_return_drop:
        return del_c

    r = df.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    return unpack_tuple(res)


def select_features_by_psi(base, no_base, target='target', threshold=0.05, include_cols=[],
                           return_drop=False, only_return_drop=False, if_select_flow=False, feature_bin=None):
    """
    通过psi筛选特征
    Args:
        base (DataFrame): 基准数据集
        no_base (DataFrame): 非基准数据集
        target (str): 目标变量名称
        threshold (float): psi筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流
        feature_bin (autobmt.FeatureBin): 特征分箱对象

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if include_cols:
        cols = include_cols
    else:
        cols = list(base.columns)

    if feature_bin is None or feature_bin == False:
        log.info('未进行分箱计算的psi')
        psi_series = psi(no_base[cols], base[cols])
    else:
        if isinstance(feature_bin, FeatureBin):
            for i in cols:  # TODO 待优化，考虑多线程
                if i not in feature_bin.splits_dict:
                    t = base[i]
                    if is_continuous(t):
                        feature_bin.fit(t, base[target], is_need_monotonic=False)
        else:
            feature_bin = FeatureBin()
            to_bin_cols = []
            for i in cols:  # TODO 待优化，考虑多线程
                if is_continuous(base[i]):
                    to_bin_cols.append(i)

            feature_bin.fit(base[to_bin_cols], base[target], is_need_monotonic=False)

        psi_series = psi(feature_bin.transform(no_base[cols]), feature_bin.transform(base[cols]))

    del_c = list(psi_series[psi_series > threshold].index)  # 大于psi阈值的变量删除

    if if_select_flow:
        return (
            del_c, threshold,
            pd.DataFrame({'feature': psi_series.index, 'PSI': psi_series}))

    if only_return_drop:
        return del_c

    r = base.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    return unpack_tuple(res)


def select_features_by_iv(df, target='target', threshold=0.02, include_cols=[], return_drop=False, return_iv=False,
                          only_return_drop=False, if_select_flow=False, feature_bin=None):
    """
    通过iv筛选特征
    Args:
        df (DataFrame): 需要进行特征筛选的数据集
        target (str): 目标变量名称
        threshold (float): iv筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        return_iv (bool): 是否返回iv
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流
        feature_bin (autobmt.FeatureBin): 特征分箱对象

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if include_cols:
        cols = np.array(include_cols)
    else:
        cols = np.array(df.columns)

    iv = np.zeros(len(cols))
    for i in range(len(cols)):
        iv[i] = calc_iv(df[cols[i]], df[target], feature_bin=feature_bin)  # 使用和select_features_by_iv_diff一样的逻辑

    drop_index = np.where(iv < threshold)

    del_c = cols[drop_index]

    if if_select_flow:
        return (del_c, threshold, pd.DataFrame({'feature': cols, 'IV': iv}))

    if only_return_drop:
        return del_c

    r = df.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    if return_iv:
        res += (pd.Series(iv, index=cols),)

    return unpack_tuple(res)


def select_features_by_iv_diff(dev, no_dev, target='target', threshold=2, include_cols=[], return_drop=False,
                               return_iv=False, only_return_drop=False, if_select_flow=False, feature_bin=None):
    """
    通过iv差值筛选特征
    Args:
        dev (DataFrame): 基准数据集
        no_dev (DataFrame): 非基准数据集
        target (str): 目标变量名称
        threshold (float): iv差值筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        return_iv (bool): 是否返回iv
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流
        feature_bin (autobmt.FeatureBin): 特征分箱对象

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if include_cols:
        cols = np.array(include_cols)
    else:
        cols = np.array(dev.columns)

    iv = np.zeros(len(cols))
    iv_no_dev = np.zeros(len(cols))
    if feature_bin is None:
        feature_bin = FeatureBin()
        for i in range(len(cols)):
            iv[i] = calc_iv(dev[cols[i]], dev[target], feature_bin=feature_bin)
            no_dev_feature = feature_bin.transform(no_dev[cols[i]])
            iv_no_dev[i] = calc_iv(no_dev_feature, no_dev[target], feature_bin=feature_bin)
    else:
        for i in range(len(cols)):
            iv[i] = calc_iv(dev[cols[i]], dev[target], feature_bin=feature_bin)
            no_dev_feature = feature_bin.transform(no_dev[cols[i]])
            iv_no_dev[i] = calc_iv(no_dev_feature, no_dev[target], feature_bin=feature_bin)

    # iv_diff = abs(iv - iv_no_dev) * 10
    iv_diff = (iv - iv_no_dev) * 10

    drop_index = np.where(iv_diff > threshold)  # IV差值大于2个点的变量剔除

    del_c = cols[drop_index]

    if if_select_flow:
        return (del_c, threshold,
                pd.DataFrame({'feature': cols, 'dev_nodev_iv_diff': iv_diff}))

    if only_return_drop:
        return del_c

    r = dev.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    if return_iv:
        res += (pd.Series(iv, index=cols),)

    return unpack_tuple(res)


def select_features_by_corr(df, target='target', by='IV', threshold=0.7, include_cols=[], return_drop=False,
                            only_return_drop=False, if_select_flow=False, feature_bin=None):
    """
    通过通过相关性筛选特征
    Args:
        df (DataFrame): 需要进行特征筛选的数据集
        target (str): 目标变量名称
        by (str|array): 用于删除特征的特征权重
        threshold (float): iv筛选阀值
        include_cols (list): 需要进行筛选的特征
        return_drop (bool): 是否返回删除的特征列表
        only_return_drop (bool): 是否只返回删除的特征列表
        if_select_flow (bool): 是否是筛选工作流
        feature_bin (autobmt.FeatureBin): 特征分箱对象

    Returns:
        DataFrame: 筛选后的数据集
        list: 删除的特征列表
    """
    if include_cols:
        cols = include_cols
    else:
        cols = list(df.columns)

    if isinstance(by, pd.DataFrame):
        by = pd.Series(by.iloc[:, 1].values, index=by.iloc[:, 0].values)

    if not isinstance(by, (str, pd.Series)):
        by = pd.Series(by, index=df.columns)

    # 计算iv
    if isinstance(by, str):
        df_corr = df[cols].corr().abs()
        # df_corr = df[cols].fillna(-999).corr().abs()
        ix, cn = np.where(np.triu(df_corr.values, 1) > threshold)  # ix是行，cn是列
        if len(ix):
            gt_thre = np.unique(np.concatenate((ix, cn)))
            gt_thre_cols = df_corr.index[gt_thre]
            iv = {}
            for i in gt_thre_cols:
                iv[i] = calc_iv(df[i], target=df[target], feature_bin=feature_bin)

            by = pd.Series(iv, index=gt_thre_cols)

    # 给重要性排下序，倒序
    by = by[list(set(by.index) & set(cols))].sort_values(ascending=False)

    by.index = by.index.astype(type(list(df.columns)[0]))
    df_corr = df[list(by.index)].corr().abs()
    # df_corr = df[list(by.index)].fillna(-999).corr().abs()

    ix, cn = np.where(np.triu(df_corr.values, 1) > threshold)

    del_all = []

    if len(ix):

        for i in df_corr:

            if i not in del_all:
                # 找出与当前特征的相关性大于域值的特征
                del_tmp = list(df_corr[i][(df_corr[i] > threshold) & (df_corr[i] != 1)].index)

                # 比较当前特征与需要删除的特征的特征重要性
                if del_tmp:
                    by_tmp = by.loc[del_tmp]
                    del_l = list(by_tmp[by_tmp <= by.loc[i]].index)
                    del_all.extend(del_l)

    del_c = list(set(del_all))

    if if_select_flow:
        return (del_c, threshold, pd.DataFrame({'feature': df_corr.index}))

    if only_return_drop:
        return del_c

    r = df.drop(columns=del_c)

    res = (r,)
    if return_drop:
        res += (del_c,)

    return unpack_tuple(res)


from .logger_utils import Logger

log = Logger(level='info', name=__name__).logger


class FeatureSelection:
    def __init__(self, df, target='target', data_type='type',
                 exclude_columns=['key', 'target', 'apply_time', 'type'], params=None,
                 match_dict=None):
        """
        特征选择模块，初始化方法
        Args:
            df (DataFrame): 需要进行变量筛选的数据集
            target (str): 目标值y变量名称
            data_type (str): 数据集划分标示的名称【即划分train、test、oot的字段名称】
            exclude_columns (list): 需要排除的特征
            match_dict (DataFrame): 数据源特征字典
            params (dict): 筛选特征的方法字典，有'empty'、'const'、'psi'、'iv'、'iv_diff'、'corr' 6种筛选方法供选择。字典形如：
            {
                'empty': {'threshold': 0.9},    #若特征的缺失值大于0.9被删除
                'const': {'threshold': 0.95},   #若特征单个值的占比大于0.95被删除
                'psi': {'threshold': 0.05},  #若特征在train、test上psi值大于0.05被删除
                'iv': {'threshold': 0.02},  #若特征的iv值小于0.02被删除
                'iv_diff': {'threshold': 2},    #若特征在train、test上的iv差值乘10后，大于2，特征被删除
                'corr': {'threshold': 0.7}, #若两个特征相关性高于0.7时，iv值低的特征被删除
            }
        """
        self.df = df
        self.target = target
        self.data_type = data_type
        self.exclude_columns = exclude_columns + [self.target]
        self.match_dict = match_dict
        self.params = params
        self.check()
        self.features = [name for name in list(self.df.columns) if name not in self.exclude_columns]
        self.feature_dict = self.get_feature_dict()
        self.select_log_df = self.build_select_log_df()
        self.step_evaluate_log_df = []  # 记录每一步的评估结果

    @property
    def get_features(self):
        """返回当前数据集最新的特征"""
        return [name for name in list(self.df.columns) if name not in self.exclude_columns]

    @property
    def get_evaluate_df_log(self):
        """合并每一步的评估结果"""
        if len(self.step_evaluate_log_df) == 0:
            log.info("并未进行评估过!!!")
            return None
        else:
            evaluate_log_df = pd.concat(self.step_evaluate_log_df, axis=0).reset_index(drop=True)
            return evaluate_log_df

    def get_feature_dict(self):
        """通过数据源简称去数据字典中获取数据源的特征名称"""
        if self.match_dict is not None and isinstance(self.match_dict, dict):
            model_name_dict = {feature: self.match_dict.get(feature.lower(), "NOT FOUNT") for feature in self.features}
            model_name_dict_df = pd.DataFrame(
                {'feature': list(model_name_dict.keys()), 'cn': list(model_name_dict.values())})
        else:
            model_name_dict_df = pd.DataFrame(
                {'feature': self.features, 'cn': ""})
        return model_name_dict_df

    def build_select_log_df(self):
        """返回特征字典，如果需要个性化修改，可以在此方法中修改"""
        return self.feature_dict

    def mapping_selection_func(self):
        """
        特征选择方法映射类
        如果需要增加新的特征选择方法，只需要增加这个字典即可
        注意:python3.6的字典默认会按照顺序进行遍历
        """
        return {
            "empty": select_features_by_miss,
            "const": select_features_by_concentrate,
            "psi": select_features_by_psi,
            "iv": select_features_by_iv,
            "iv_diff": select_features_by_iv_diff,
            "corr": select_features_by_corr,
        }

    def select(self):
        """
        执行定义的特征选择方法，返回筛选过后的数据集，剩余特征名称，以及筛选过程
        Returns:

        """
        log.info('开始执行特征选择模块... 数据集结构为[{}]'.format(self.df.shape))

        if self.params is None:  # 默认只进行3种方式进行变量筛选
            self.params = {
                'empty': {'threshold': 0.9},
                # 'const': {'threshold': 0.95},
                # 'psi': {'threshold': 0.05},
                'iv': {'threshold': 0.02},
                # 'iv_diff': {'threshold': 2},
                'corr': {'threshold': 0.7},
            }
            if self.data_type in self.df:
                self.params['psi'] = {'threshold': 0.05, 'target': self.target}
            log.info('未指定筛选方法的阈值，使用默认方法和阈值：{}'.format(self.params))

        if self.data_type in self.df:
            dev_data = self.df[self.df[self.data_type] == 'train']
            if 'oot' in np.unique(self.df[self.data_type]):
                nodev_data = self.df[self.df[self.data_type] == 'oot']
            else:
                nodev_data = self.df[self.df[self.data_type] == 'test']
        else:
            dev_data = self.df

        fb = FeatureBin()

        for k, v in self.params.items():
            v.update({'if_select_flow': True})
            v.update({'include_cols': self.get_features})
            if k in ['iv', 'iv_diff', 'psi', 'corr']:
                v.update({'feature_bin': fb})
                if 'target' not in v:
                    v.update({'target': self.target})
            # 执行具体的筛选方法
            if k in ['iv_diff', 'psi'] and self.data_type in self.df:
                del_c, th, fea_value_df = self.mapping_selection_func()[k](dev_data, nodev_data, **v)
                log.info('删除变量 ：{}'.format(del_c))
            else:
                del_c, th, fea_value_df = self.mapping_selection_func()[k](dev_data, **v)
                log.info('删除变量 ：{}'.format(del_c))
            if k == 'iv':
                # 将算好的iv值放进去
                self.params['corr']['by'] = fea_value_df

            self.df = self.df.drop(columns=del_c)

            self.select_log_df = self.select_log_df.merge(fea_value_df, on='feature', how='left')
            if k == 'iv':
                step_name = "{}_selection_feature_flag (<{})".format(k, th)
            else:
                step_name = "{}_selection_feature_flag (>{})".format(k, th)
            log.info("{} 方法剔除变量，阈值为{}，剔除的变量有 : {} 个".format(k, step_name, len(del_c)))
            self.select_log_df[step_name] = self.select_log_df['feature'].map(lambda x: 1 if x in del_c else 0)

        # 所有的特征选择方法，只要命中一个，即剔除该特征
        filter_condition_features = [name for name in list(self.select_log_df.columns) if '_feature_flag' in name]
        self.select_log_df['feature_filter_flag'] = self.select_log_df[filter_condition_features].sum(axis=1)
        log.info('特征选择执行完成... 数据集结构为[{}]'.format(self.df.shape))

        # return self.df, self.get_features, self.select_log_df, self.get_evaluate_df_log, fb
        return self.df, self.get_features, self.select_log_df, fb

    def check(self):
        """
        特征选择模块，前置检查，符合要求，则往下运行
        Returns:

        """
        log.info('开始进行前置检查')
        if self.df is None or not isinstance(self.df, pd.DataFrame):
            raise ValueError('数据集不能为空并且数据集必须是dataframe!!!')

        if self.data_type not in self.df:
            log.info('train、test数据集标识的字段名不存在！或未进行数据集的划分，筛选变量无法使用psi、iv_diff进行筛选，建议请将数据集划分为train、test!!!')
            if self.params is not None:
                if 'psi' in self.params:
                    del self.params['psi']
                if 'iv_diff' in self.params:
                    del self.params['iv_diff']
        else:
            data_type_ar = np.unique(self.df[self.data_type])
            if 'train' not in data_type_ar:
                raise KeyError("""没有开发样本，数据集标识字段{}没有`train`该取值!!!""".format(self.data_type))

            if 'test' not in data_type_ar:
                raise KeyError("""没有验证样本，数据集标识字段{}没有`test`该取值!!!""".format(self.data_type))

        if self.target is None:
            raise ValueError('数据集的目标变量名称不能为空!!!')

        if self.target not in self.df:
            raise KeyError('样本中没有目标变量y值!!!')

        if self.exclude_columns is None or self.target not in self.exclude_columns:
            raise ValueError('exclude_columns 不能为空，必须包含target字段!!!')
        n_cols, c_cols, d_cols = select_features_dtypes(self.df, exclude=self.exclude_columns)
        log.info('数值特征个数: {}'.format(len(n_cols)))
        log.info('字符特征个数: {}'.format(len(c_cols)))
        log.info('日期特征个数: {}'.format(len(d_cols)))
        if len(c_cols) > 0:
            log.info('数据集中包含有{}个字符特征,{}个日期特征'.format(len(c_cols), len(d_cols)))
